<?php

namespace QUI\InviteCode;

use DateTime;
use QUI;
use QUI\Database\Exception;
use QUI\InviteCode\Exception\InviteCodeException;
use QUI\InviteCode\Exception\InviteCodeMailException;
use QUI\Permissions\Permission;

/**
 * Class InviteCode
 */
class InviteCode
{
    /**
     * InviteCode ID
     *
     * @var int
     */
    protected int $id;

    /**
     * Actual code
     *
     * @var string
     */
    protected mixed $code;

    /**
     * User that is assigned to this Code
     *
     * @var QUI\Interfaces\Users\User | null
     */
    protected null | QUI\Interfaces\Users\User $User = null;

    /**
     * Email address that is asasigned to this Code
     *
     * @var string
     */
    protected mixed $email = null;

    /**
     * Creation Date
     *
     * @var DateTime
     */
    protected DateTime $CreateDate;

    /**
     * Use Date
     *
     * @var null | DateTime
     */
    protected ?DateTime $UseDate = null;

    /**
     * Date until the Invite code is valid
     *
     * @var null | DateTime
     */
    protected ?DateTime $ValidUntilDate = null;

    /**
     * InviteCode title
     *
     * @var string
     */
    protected mixed $title;

    /**
     * Flag if mail has been sent
     *
     * @var bool
     */
    protected bool $mailSent;

    /**
     * @var bool
     */
    protected bool $valid = true;

    /**
     * InviteCode constructor.
     *
     * @param int $id - Invite Code ID
     *
     * @throws InviteCodeException
     * @throws Exception
     * @throws \DateMalformedStringException
     */
    public function __construct(int $id)
    {
        $result = QUI::getDataBase()->fetch([
            'from' => Handler::getTable(),
            'where' => [
                'id' => $id
            ]
        ]);

        if (empty($result)) {
            throw new InviteCodeException([
                'quiqqer/invitecode',
                'exception.invitecode.not_found',
                [
                    'id' => $id
                ]
            ], 404);
        }

        $data = current($result);

        $this->id = (int)$data['id'];
        $this->code = $data['code'];
        $this->title = $data['title'];
        $this->mailSent = boolval($data['mailSent']);

        if (!empty($data['userId'])) {
            try {
                $this->User = QUI::getUsers()->get($data['userId']);
            } catch (\Exception $Exception) {
//                QUI\System\Log::addWarning(
//                    'Could not find User #' . $data['userId'] . ' for Invite Code #' . $this->id . '.'
//                );
//
//                QUI\System\Log::writeException($Exception);
            }
        }

        if (!empty($data['email'])) {
            $this->email = $data['email'];
        }

        $this->CreateDate = new DateTime($data['createDate']);

        if (!empty($data['useDate'])) {
            $this->UseDate = new DateTime($data['useDate']);
        }

        if (!empty($data['validUntilDate'])) {
            $this->ValidUntilDate = new DateTime($data['validUntilDate']);

            $Now = new DateTime();

            if (!$this->isRedeemed() && $Now > $this->ValidUntilDate) {
                $this->valid = false;
            }
        }
    }

    /**
     * @return int
     */
    public function getId(): int
    {
        return $this->id;
    }

    /**
     * @return string
     */
    public function getCode(): string
    {
        return $this->code;
    }

    /**
     * @return QUI\Interfaces\Users\User|null
     */
    public function getUser(): ?QUI\Interfaces\Users\User
    {
        return $this->User;
    }

    /**
     * @param QUI\Interfaces\Users\User $User
     */
    public function setUser(QUI\Interfaces\Users\User $User): void
    {
        $this->User = $User;
    }

    /**
     * @return string
     */
    public function getEmail(): ?string
    {
        return $this->email;
    }

    /**
     * @param string $email
     */
    public function setEmail(string $email): void
    {
        $this->email = $email;
    }

    /**
     * @return DateTime
     */
    public function getCreateDate(): DateTime
    {
        return $this->CreateDate;
    }

    /**
     * @return DateTime|null
     */
    public function getUseDate(): ?DateTime
    {
        return $this->UseDate;
    }

    /**
     * @return DateTime|null
     */
    public function getValidUntilDate(): ?DateTime
    {
        return $this->ValidUntilDate;
    }

    /**
     * @return string
     */
    public function getTitle(): string
    {
        return $this->title;
    }

    /**
     * @return bool
     */
    public function isMailSent(): bool
    {
        return $this->mailSent;
    }

    /**
     * Redeems this InviteCode
     *
     * Hint: This invalidates the code for future use
     *
     * @param QUI\Interfaces\Users\User $User
     * @return void
     * @throws InviteCodeException
     * @throws QUI\Exception
     */
    public function redeem(QUI\Interfaces\Users\User $User): void
    {
        if ($this->isRedeemed()) {
            throw new InviteCodeException([
                'quiqqer/invitecode',
                'exception.invitecode.already_used'
            ]);
        }

        if (!$this->isValid()) {
            throw new InviteCodeException([
                'quiqqer/invitecode',
                'exception.invitecode.no_longer_valid'
            ]);
        }

        $Now = new DateTime();

        QUI::getDataBase()->update(
            Handler::getTable(),
            [
                'useDate' => $Now->format('Y-m-d H:i:s'),
                'userId' => $User->getUUID()
            ],
            [
                'id' => $this->id
            ]
        );

        $this->UseDate = $Now;

        QUI::getEvents()->fireEvent('quiqqerInviteCodeRedeem', [$this, $User]);
    }

    /**
     * Check if this InviteCode is still valid
     *
     * @return bool
     */
    public function isValid(): bool
    {
        return $this->valid;
    }

    /**
     * Checks if an InviteCode is redeemed
     *
     * @return bool
     */
    public function isRedeemed(): bool
    {
        return !is_null($this->getUseDate());
    }

    /**
     * Send this Invite Code via E-Mail
     *
     * @param bool $resend (optional) - Send again if already send [default: false]
     * @return void
     *
     * @throws Exception
     * @throws InviteCodeMailException
     * @throws QUI\Exception
     * @throws \PHPMailer\PHPMailer\Exception
     */
    public function sendViaMail(bool $resend = false): void
    {
        if (!$resend && $this->isMailSent()) {
            return;
        }

        $email = $this->getEmail();

        if (empty($email)) {
            throw new InviteCodeMailException([
                'quiqqer/invitecode',
                'exception.invitecode.no_email'
            ]);
        }

        $Mailer = new QUI\Mail\Mailer();

        $Mailer->addRecipient($email);

        $Engine = QUI::getTemplateManager()->getEngine();
        $dir = QUI::getPackage('quiqqer/invitecode')->getDir() . 'templates/';
        $validUntilDate = $this->getValidUntilDate();
        $data = [
            'code' => $this->getCode(),
            'validUntilDate' => empty($validUntilDate) ? '' : QUI::getLocale()->formatDate(
                $validUntilDate->getTimestamp()
            )
        ];

        $RegistrationSite = Handler::getRegistrationSite();

        if (empty($RegistrationSite)) {
            throw new InviteCodeMailException([
                'quiqqer/invitecode',
                'exception.invitecode.no_registration_site'
            ]);
        }

        $data['registrationUrl'] = $RegistrationSite->getUrlRewrittenWithHost();
        $data['email'] = $email;
        $data['inviteCode'] = $this->getCode();

        if (empty($validUntilDate)) {
            $translationVar = 'mail.invite_code.body';
        } else {
            $translationVar = 'mail.invite_code.body_date';
        }

        $Engine->assign([
            'body' => QUI::getLocale()->get(
                'quiqqer/invitecode',
                $translationVar,
                $data
            )
        ]);

        $Mailer->setSubject(
            QUI::getLocale()->get(
                'quiqqer/invitecode',
                'mail.invite_code.subject'
            )
        );

        $Mailer->setBody($Engine->fetch($dir . 'mail.invite_code.html'));
        $Mailer->send();

        // update internal flag
        QUI::getDataBase()->update(
            Handler::getTable(),
            [
                'mailSent' => 1
            ],
            [
                'id' => $this->getId()
            ]
        );
    }

    /**
     * Permanently delete this InviteCode
     *
     * @return void
     * @throws QUI\Permissions\Exception|Exception
     */
    public function delete(): void
    {
        Permission::checkPermission(Handler::PERMISSION_DELETE);

        QUI::getDataBase()->delete(
            Handler::getTable(),
            [
                'id' => $this->id
            ]
        );
    }

    /**
     * Get InviteCode attributes as array
     *
     * @return array
     */
    public function toArray(): array
    {
        $data = [
            'id' => $this->getId(),
            'code' => $this->getCode(),
            'userId' => false,
            'username' => false,
            'email' => $this->getEmail() ?: false,
            'createDate' => $this->getCreateDate()->format('Y-m-d H:i:s'),
            'useDate' => false,
            'validUntilDate' => false,
            'title' => $this->getTitle() ?: false,
            'mailSent' => $this->isMailSent(),
            'valid' => $this->isValid()
        ];

        $User = $this->getUser();

        if ($User) {
            $data['userId'] = $User->getUUID();
            $data['username'] = $User->getName();
        }

        $UseDate = $this->getUseDate();

        if ($UseDate) {
            $data['useDate'] = $UseDate->format('Y-m-d H:i:s');
        }

        $ValidUntilDate = $this->getValidUntilDate();

        if ($ValidUntilDate) {
            $data['validUntilDate'] = $ValidUntilDate->format('Y-m-d H:i:s');
        }

        return $data;
    }
}
